from abc import ABC

from multipledispatch import dispatch

from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.core.ordinal_ignore_case_dict import OrdinalIgnoreCaseDict
import pyconfig_extension.abstracts.configuration_path as conf_path


class ConfigurationProvider(IConfigurationProvider, ABC):
    def __init__(self):
        self._data = OrdinalIgnoreCaseDict()

    def try_get(self, key: str) -> tuple[bool, str | None]:
        return self._data.try_get(key)

    def set(self, key: str, value: str):
        self._data[key] = value

    def load(self):
        pass

    def get_child_keys(self, earlier_keys: list[str], parent_path: str | None):
        results: list[str] = []

        if parent_path is None:
            for key in self._data:
                results.append(self.__segment(key, 0))
        else:
            assert conf_path.KEY_DELIMITER == ":"
            for key in self._data:
                if len(key) > len(parent_path) and \
                        key.lower().startswith(parent_path.lower()) and \
                        key[len(parent_path)] == conf_path.KEY_DELIMITER:
                    results.append(self.__segment(key, len(parent_path) + 1))

        results.extend(earlier_keys)
        results.sort()
        return results

    @staticmethod
    def __segment(key: str, prefix_length) -> str:
        try:
            index = key.index(conf_path.KEY_DELIMITER, prefix_length)
            return key[prefix_length: index]
        except ValueError:
            return key[prefix_length:]

    def __str__(self):
        return type(self).__name__



from abc import abstractmethod

from pyconfig_extension import guard
from pyconfig_extension.core.configuration_provider import ConfigurationProvider
from pyconfig_extension.file.file_configuration_source import FileConfigurationSource


class FileConfigurationProvider(ConfigurationProvider):

    def __init__(self, source: FileConfigurationSource):
        super().__init__()
        guard.check_none(source)
        self._source = source

    @property
    def source(self) -> FileConfigurationSource:
        return self._source

    def load(self):
        if self.source.stream is not None:
            self.load_stream(self.source.stream)
        elif self.source.path is not None and self.source.stream is None:
            try:
                with open(self.source.path, "r") as f:
                    content = f.read()
                    self.source.stream = str(content)
                    self.load_stream(self.source.stream)
            except Exception as ex:
                if not self.source.optional:
                    raise ex

    @abstractmethod
    def load_stream(self, stream):
        raise NotImplementedError

    def __str__(self):
        return f'{type(self).__name__} for {self._source.path} ({"Optional" if self._source.optional else "Required"})'



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.stream_configuration_source import StreamConfigurationSource


class JsonStreamConfigurationSource(StreamConfigurationSource):
    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.json.json_stream_configuration_provider import JsonStreamConfigurationProvider
        return JsonStreamConfigurationProvider(self)



from functools import cmp_to_key

from pyconfig_extension.abstracts.iconfiguration import IConfigurationSection
from pyconfig_extension.abstracts.iconfiguration_root import IConfigurationRoot
import pyconfig_extension.abstracts.configuration_path as path_conf
from pyconfig_extension.core.configuration_key_comparer import compare


def get_children_implementation(root: IConfigurationRoot, path: str | None) -> list[IConfigurationSection]:
    providers = root.providers
    if not any(providers):
        return []

    children_keys = providers[0].get_child_keys([], path)

    for i in range(1, len(providers)):
        children_keys = providers[i].get_child_keys(children_keys, path)

    children_keys.sort(key=cmp_to_key(compare))
    # ignore case distinct
    dict_no_case: dict[str, str] = {}
    for children_key in children_keys:
        if children_key.lower() not in dict_no_case:
            dict_no_case[children_key.lower()] = children_key

    list_no_case = list(dict_no_case.values())

    children: list[IConfigurationSection] = []
    for key in list_no_case:
        section_key = key if path is None else path_conf.combine(path, key)
        child = root.get_section(section_key)
        children.append(child)
    return children



from typing import Callable

from multipledispatch import dispatch

from pyconfig_extension import guard
from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.abstracts.configuration_extensions import add
from pyconfig_extension.ini.ini_configuration_source import IniConfigurationSource
from pyconfig_extension.ini.ini_stream_configuration_source import IniStreamConfigurationSource


def add_ini_string(builder: IConfigurationBuilder, stream: str) -> IConfigurationBuilder:
    def config_function(x: IniStreamConfigurationSource):
        x.stream = stream

    return add(builder, IniStreamConfigurationSource, config_function)


@dispatch(IConfigurationBuilder, str)
def add_ini_file(builder: IConfigurationBuilder, path: str) -> IConfigurationBuilder:
    return add_ini_file(builder, path, True)


@dispatch(IConfigurationBuilder, object)
def add_ini_file(builder: IConfigurationBuilder, configure_source: Callable) -> IConfigurationBuilder:
    return add(builder, IniConfigurationSource, configure_source)


@dispatch(IConfigurationBuilder, str, bool)
def add_ini_file(builder: IConfigurationBuilder, path: str, optional: bool) -> IConfigurationBuilder:
    guard.check_none(path)

    def config_source(x: IniConfigurationSource):
        x.path = path
        x.optional = optional

    return add_ini_file(builder, config_source)



from abc import ABC, abstractmethod
from typing import Callable

from multipledispatch import dispatch

from pyconfig_extension.abstracts.binder.binder_option_listener import IBinderOptionConfigure, DefaultBinderOptionConfigure
from pyconfig_extension.abstracts.binder.binder_options import BinderOptions
from pyconfig_extension.abstracts.binder.binding_point import BindingPoint


class IConfiguration(ABC):

    @abstractmethod
    def __getitem__(self, item: str) -> str | None:
        raise NotImplementedError

    @abstractmethod
    def __setitem__(self, key: str, value: str | None):
        raise NotImplementedError

    @abstractmethod
    def get_section(self, key: str) -> 'IConfigurationSection':
        raise NotImplementedError

    @abstractmethod
    def get_children(self) -> list['IConfigurationSection']:
        raise NotImplementedError

    def get_required_section(self, key: str):
        section = self.get_section(key)
        if section is not None and section.exists():
            return section
        raise Exception(f"section {key} is not found in configuration")

    @dispatch(object)
    def get(self, klass_type: type):
        def binder_function(x: BinderOptions):
            pass

        return self.get(klass_type, binder_function)

    @dispatch(str, object)
    def get(self, section_key: str, klass_type: type):
        def binder_function(x: BinderOptions):
            pass

        section = self.get_section(section_key)
        return section.get(klass_type, binder_function)

    @dispatch(object, object)
    def get(self, klass_type: type, configure_options: Callable[[BinderOptions], None]):
        options = BinderOptions()
        configure_options(options)

        from pyconfig_extension.abstracts.binder.binder_utils import get
        return get(self, klass_type, configure_options)

    @dispatch(str, object, object)
    def bind(self, section_key: str, instance, instance_typ: type):
        section: IConfigurationSection = self.get_section(section_key)
        section.bind(instance, instance_typ, DefaultBinderOptionConfigure())

    @dispatch(str, object)
    def bind(self, section_key: str, instance):
        section: IConfigurationSection = self.get_section(section_key)
        section.bind(instance, type(instance), DefaultBinderOptionConfigure())

    @dispatch(object, object)
    def bind(self, instance, instance_type: type):
        self.bind(instance, instance_type, DefaultBinderOptionConfigure())

    @dispatch(object)
    def bind(self, instance):
        self.bind(instance, type(instance), DefaultBinderOptionConfigure())

    @dispatch(object, object, IBinderOptionConfigure)
    def bind(self, instance, instance_type: type, configure_options: IBinderOptionConfigure):

        if instance is not None:
            options = BinderOptions()
            if configure_options is not None:
                configure_options.on_config(options)
            binding_point = BindingPoint(instance, True)

            from pyconfig_extension.abstracts.binder.binder_utils import bind_instance
            bind_instance(instance_type, binding_point, self, options)


class IConfigurationSection(IConfiguration, ABC):

    @property
    @abstractmethod
    def key(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def path(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def value(self) -> str | None:
        raise NotImplementedError

    def exists(self):
        if self is None:
            return False
        return self.value is not None or len(self.get_children()) > 0



from pyconfig_extension import guard

KEY_DELIMITER: str = ":"


def combine(*path_segments: str):
    guard.check_none(path_segments)
    res = ':'.join(path_segments)
    return res


def get_section_key(path: str | None):
    if path is None or len(path) == 0:
        return path

    try:
        index = path.rindex(KEY_DELIMITER)
        return path[index + 1:]
    except ValueError:
        return path


def get_parent_path(path: str | None):
    if path is None or len(path) == 0:
        return None

    try:
        index = path.rindex(KEY_DELIMITER)
        return path[:index]
    except ValueError:
        return None



from pyconfig_extension.core.configuration_provider import ConfigurationProvider
from pyconfig_extension.core.memory_configuration_source import MemoryConfigurationSource


class MemoryConfigurationProvider(ConfigurationProvider):
    def __init__(self, source: MemoryConfigurationSource):
        super().__init__()
        self.__source = source
        if self.__source.initial_data is not None:
            for key in self.__source.initial_data:
                self._data[key] = self.__source.initial_data[key]




from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationSource, IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider


class CommandLineConfigurationSource(IConfigurationSource):
    switch_mappings: dict[str, str]
    args: list[str] = []

    def __init__(self, args: list[str] = None, switch_mappings: dict[str, str] = None):
        self.args = args if args is not None else []
        self.switch_mappings = switch_mappings

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.command_line.provider import CommandLineConfigurationProvider
        return CommandLineConfigurationProvider(self.args, self.switch_mappings)



from typing import Callable

from multipledispatch import dispatcher, dispatch


class BindingPoint:

    def __init__(self, initial_value=None, is_readonly: bool = False):
        self.__is_readonly = is_readonly
        self.__initial_value = None
        self._initial_value_provider = None
        if isinstance(initial_value, Callable):
            self._initial_value_provider = initial_value
        else:
            self.__initial_value = initial_value
        self._is_value_set = False
        self._set_value = None

    @property
    def is_readonly(self) -> bool:
        return self.__is_readonly

    @property
    def has_new_value(self):
        if self.is_readonly:
            return False
        return self._is_value_set

    @property
    def value(self):
        if self._is_value_set:
            return self._set_value
        else:
            if self.__initial_value is not None:
                return self.__initial_value
            elif self._initial_value_provider is not None:
                return self._initial_value_provider()
            return None

    def set_value(self, new_value):
        assert not self.is_readonly
        assert not self._is_value_set
        self._is_value_set = True
        self._set_value = new_value

    def try_set_value(self, new_value):
        if not self.is_readonly:
            self.set_value(new_value)



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationSource, IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider


class EnvironmentVariablesConfigurationSource(IConfigurationSource):
    prefix: str | None

    def __init__(self, prefix: str | None = None):
        self.prefix = prefix

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.environment_variables.provider import EnvironmentVariablesConfigurationProvider
        return EnvironmentVariablesConfigurationProvider(self.prefix)



from pyconfig_extension import guard
from pyconfig_extension.core.configuration_provider import ConfigurationProvider
from pyconfig_extension.core.ordinal_ignore_case_dict import OrdinalIgnoreCaseDict


def index_of(value: str, pattern: str):
    try:
        return value.index(pattern)
    except:
        return -1


class CommandLineConfigurationProvider(ConfigurationProvider):

    def __init__(self, args: list[str], switch_mappings: dict[str, str] = None):
        guard.check_none(args)
        super().__init__()
        self._args = args
        self._switch_mappings = switch_mappings
        if self._switch_mappings is not None:
            self._switch_mappings = self.__get_validated_switch_mappings(switch_mappings)

    @property
    def args(self) -> list[str]:
        return self._args

    def load(self):
        self._data = self.__load_args(self.args, self._switch_mappings)

    @staticmethod
    def __load_args(args, switch_mapping) -> OrdinalIgnoreCaseDict:
        result = OrdinalIgnoreCaseDict()
        args_iterator = iter(args)
        for current_arg in args_iterator:
            key_start_index = 0

            if current_arg.startswith("--"):
                key_start_index = 2
            elif current_arg.startswith("-"):
                key_start_index = 1
            elif current_arg.startswith("/"):
                current_arg = f"--{current_arg[1:]}"
                key_start_index = 2

            seperator = index_of(current_arg, "=")

            if seperator < 0:
                if key_start_index == 0:
                    continue
                if switch_mapping is not None and current_arg in switch_mapping:
                    key = switch_mapping[current_arg]
                # If the switch starts with a single "-" and it isn't in given mappings ,
                # it is an invalid usage so ignore it
                elif key_start_index == 1:
                    continue
                else:
                    key = current_arg[key_start_index:]

                try:
                    value = next(args_iterator)
                except StopIteration:
                    continue
            else:
                key_segment = current_arg[:seperator]

                if switch_mapping is not None and key_segment in switch_mapping:
                    key = switch_mapping[key_segment]
                # If the switch starts with a single "-" and it isn't in given mappings ,
                # it is an invalid usage so ignore it
                elif key_start_index == 1:
                    raise Exception(f"The short switch '{current_arg}' is not defined in the switch mappings.")
                else:
                    key = current_arg[key_start_index: seperator]
                value = current_arg[seperator + 1:]
            result[key] = value

        return result

    @staticmethod
    def __get_validated_switch_mappings(switch_mappings) -> OrdinalIgnoreCaseDict:
        result = OrdinalIgnoreCaseDict()
        for mapping_key in switch_mappings:
            mapping_key: str
            if not mapping_key.startswith("-") and not mapping_key.startswith("--"):
                raise Exception(
                    f"The switch mappings contain an invalid switch '{mapping_key}' , should start with '-' or '--'.")
            if mapping_key in result:
                raise Exception(
                    f"Keys in switch mappings are case-insensitive. A duplicated key '{mapping_key}' was found.")

            result[mapping_key] = switch_mappings[mapping_key]
        return result



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder, IConfigurationSource
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider


class MemoryConfigurationSource(IConfigurationSource):

    def __init__(self, initial_data: dict[str, str] | None = None):
        self.__initial_date: dict[str, str] | None = None
        if initial_data is not None:
            self.__initial_date = initial_data

    @property
    def initial_data(self) -> dict[str, str] | None:
        return self.__initial_date

    @initial_data.setter
    def initial_data(self, value: dict[str, str] | None):
        self.__initial_date = value

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.core.memory_configuration_provider import MemoryConfigurationProvider
        return MemoryConfigurationProvider(self)



from abc import ABC

from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationSource, IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider


class FileConfigurationSource(IConfigurationSource, ABC):
    path: str
    """
    The path to the file.
    """
    # file_provider: object
    # """
    # Used to access the contents of the file.
    # """
    optional: bool
    """
    Determines if loading the file is optional.
    """

    def __init__(self, stream=None, path: str = None, optional: bool = False):
        self.stream = stream
        self.path: str = path
        self.optional = optional

    def ensure_defaults(self, builder: IConfigurationBuilder):
        pass



from abc import ABC, abstractmethod

from pyconfig_extension.abstracts.binder.binder_options import BinderOptions


class IBinderOptionConfigure(ABC):

    @abstractmethod
    def on_config(self, options: BinderOptions):
        pass


class DefaultBinderOptionConfigure(IBinderOptionConfigure):

    def on_config(self, options: BinderOptions):
        pass



from pyconfig_extension.abstracts.iconfiguration import IConfiguration, IConfigurationSection
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.core.configuration_builder import ConfigurationBuilder

import pyconfig_extension.core.extensions as core_extension
import pyconfig_extension.abstracts.configuration_extensions as abstract_extension
import pyconfig_extension.json.extensions as json_extension
import pyconfig_extension.ini.extensions as ini_extension
import pyconfig_extension.command_line.extensions as cmd_extension
import pyconfig_extension.environment_variables.extensions as env_extension


def add_json_file(*args):
    return json_extension.add_json_file(*args)


def add_json_string(*args):
    return json_extension.add_json_string(*args)


def add_ini_file(*args):
    return ini_extension.add_ini_file(*args)


def add_ini_string(*args):
    return ini_extension.add_ini_string(*args)


def add_command_line(*args):
    return cmd_extension.add_command_line(*args)


def add_in_memory_collection(*args):
    return core_extension.add_in_memory_collection(*args)


def add_configuration(*args):
    return core_extension.add_configuration(*args)


IConfigurationBuilder.add_configuration = add_configuration
IConfigurationBuilder.add_in_memory_collection = add_in_memory_collection

IConfigurationBuilder.add_ini_file = add_ini_file
IConfigurationBuilder.add_ini_string = add_ini_string

IConfigurationBuilder.add_json_file = add_json_file
IConfigurationBuilder.add_json_string = add_json_string

IConfigurationBuilder.add_command_line = add_command_line



from multipledispatch import dispatch

from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
import pyconfig_extension.abstracts.configuration_extensions as config_ext
from pyconfig_extension.json.json_configuration_source import JsonConfigurationSource
from pyconfig_extension.json.json_stream_configuration_source import JsonStreamConfigurationSource


def add_json_string(builder: IConfigurationBuilder, stream):
    def configure(x: IConfigurationBuilder):
        x.stream = stream

    return config_ext.add(builder, JsonStreamConfigurationSource, configure)


@dispatch(IConfigurationBuilder, str)
def add_json_file(builder: IConfigurationBuilder, path: str):
    return add_json_file(builder, path, False)


@dispatch(IConfigurationBuilder, str, bool)
def add_json_file(builder: IConfigurationBuilder, path: str, optional: bool):
    def configure(x: JsonConfigurationSource):
        x.optional = optional
        x.path = path

    return config_ext.add(builder, JsonConfigurationSource, configure)



from pyconfig_extension.abstracts.stream_configuration_provider import StreamConfigurationProvider
from pyconfig_extension.json.json_stream_configuration_source import JsonStreamConfigurationSource
from pyconfig_extension.json.json_configuration_file_parser import JsonConfigurationFileParser


class JsonStreamConfigurationProvider(StreamConfigurationProvider):

    def __init__(self, source: JsonStreamConfigurationSource):
        super().__init__(source)

    def load_stream(self, stream):
        self._data = JsonConfigurationFileParser.parse(stream)



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.core.extensions import add_configuration

IConfigurationBuilder.add_configuration = add_configuration



from abc import ABC, abstractmethod

from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.iconfiguration_root import IConfigurationRoot


class IConfigurationBuilder(ABC):

    @property
    @abstractmethod
    def properties(self) -> dict[str, object]:
        raise NotImplementedError

    @property
    @abstractmethod
    def sources(self) -> list['IConfigurationSource']:
        raise NotImplementedError

    @abstractmethod
    def add(self, source: 'IConfigurationSource') -> 'IConfigurationBuilder':
        raise NotImplementedError

    @abstractmethod
    def build(self) -> IConfigurationRoot:
        raise NotImplementedError


class IConfigurationSource(ABC):

    @abstractmethod
    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        raise NotImplementedError



from abc import ABC

from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationSource


class StreamConfigurationSource(IConfigurationSource, ABC):
    stream: object
    """
    The stream containing the configuration data.
    """



from pyconfig_extension.abstracts.iconfiguration import IConfiguration
from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationSource, IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider


class ChainedConfigurationSource(IConfigurationSource):

    def __init__(self, configuration: IConfiguration, should_exit_configuration: bool = False):
        self._configuration = configuration
        self._should_exit_configuration = should_exit_configuration

    @property
    def configuration(self) -> IConfiguration:
        return self._configuration

    @configuration.setter
    def configuration(self, value: IConfiguration):
        self._configuration = value

    @property
    def should_exit_configuration(self) -> bool:
        return self._should_exit_configuration

    @should_exit_configuration.setter
    def should_exit_configuration(self, value: bool):
        self._should_exit_configuration = value

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.core.chained_configuration_provider import ChainedConfigurationProvider
        return ChainedConfigurationProvider(self)



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.environment_variables.source import EnvironmentVariablesConfigurationSource


def add_environment_variables(builder: IConfigurationBuilder, prefix: str | None = None) -> IConfigurationBuilder:
    builder.add(EnvironmentVariablesConfigurationSource(prefix))
    return builder



import enum
import inspect
import logging
from types import GenericAlias
from typing import Callable

from pyconfig_extension import guard
from pyconfig_extension.abstracts.binder.binder_options import BinderOptions
from pyconfig_extension.abstracts.binder.binding_point import BindingPoint
from pyconfig_extension.abstracts.iconfiguration import IConfiguration, IConfigurationSection


def get(configuration: IConfiguration, klass_type: type, configure_options: Callable[[BinderOptions], None]) -> object:
    guard.check_none(configuration)

    options = BinderOptions()
    if configure_options is not None:
        configure_options(options)

    binding_point = BindingPoint()
    bind_instance(klass_type, binding_point, configuration, options)
    return binding_point.value


def bind_instance(klass_type: type, binding_point: BindingPoint, config: IConfiguration, options: BinderOptions):
    if klass_type == IConfigurationSection:
        binding_point.try_set_value(config)
        return
    if isinstance(config, IConfigurationSection) and config.value is not None:
        check, result, error = try_convert_value(klass_type, config.value, config.path)
        if check:
            if error is not None:
                raise error
            binding_point.try_set_value(result)
            return

    if config is not None and len(config.get_children()) > 0:
        if binding_point.value is None:
            if binding_point.is_readonly:
                return
            instance = create_instance(klass_type, config, options)
            binding_point.set_value(instance)

        instance = binding_point.value
        if _is_list(klass_type):
            bind_collection(instance, config, options, klass_type)
        elif _is_set(klass_type):
            bind_set(instance, config, options, klass_type)
        elif _is_dictionary(klass_type):
            bind_dictionary(instance, config, options, klass_type)
        else:
            bind_properties(instance, config, options)


def _is_list(klass_type: type):
    return issubclass(klass_type, list) or \
        (isinstance(klass_type, GenericAlias) and issubclass(klass_type.__origin__, list))


def _is_set(klass_type: type):
    return issubclass(klass_type, set) or \
        (isinstance(klass_type, GenericAlias) and issubclass(klass_type.__origin__, set))


def _is_dictionary(klass_type: type):
    return issubclass(klass_type, dict) or \
        (isinstance(klass_type, GenericAlias) and issubclass(klass_type.__origin__, dict))


def create_instance(klass_type: type, config: IConfiguration, options: BinderOptions):
    return klass_type()


def try_convert_value(klass_type: type, value: str, path: str) -> tuple[bool, object, Exception]:
    error = None
    result = None
    if klass_type == object:
        return True, value, error
    if klass_type == str:
        return True, value, error
    if klass_type == bool:
        if value.lower() == "true":
            return True, True, error
        elif value.lower() == "false":
            return True, False, error
        else:
            error = ValueError(f"invalid value format for value = {value} , path = {path}")
            return False, None, error
    val = klass_type(value)
    return True, val, error


def bind_properties(instance, configuration: IConfiguration, options: BinderOptions):
    annotations = get_all_properties(type(instance))
    for property_name in annotations:
        property_type = annotations[property_name]
        bind_property(property_name, property_type, instance, configuration, options)


def bind_property(property_name: str, property_type, instance, config: IConfiguration, options: BinderOptions):
    def initial_provider_for_property():
        if hasattr(instance, property_name):
            return getattr(instance, property_name)

    property_binding_point = BindingPoint(initial_provider_for_property, is_readonly=False)
    section = config.get_section(property_name)
    bind_instance(property_type, property_binding_point, section, options)
    if property_binding_point.value is not None:
        val = property_binding_point.value
        setattr(instance, property_name, val)


def get_all_properties(klass_type: type):
    mro = inspect.getmro(klass_type)
    annotations = {}
    for cls in mro:
        class_annotations = inspect.get_annotations(cls)
        annotations.update(class_annotations)
    res = {k: v for k, v in annotations.items() if not k.startswith("_")}
    return res


def get_collection_arg_type(klass_type):
    if isinstance(klass_type, GenericAlias):
        return klass_type.__args__[0]
    if issubclass(klass_type, list) and hasattr(klass_type, "__orig_bases__"):
        return get_collection_arg_type(klass_type.__orig_bases__[0])
    raise NotImplementedError


def bind_collection(collection: list, config: IConfiguration, options: BinderOptions, klass_type: type):
    arg_type = get_collection_arg_type(klass_type)
    for section in config.get_children():
        try:
            item_binding_point = BindingPoint()
            bind_instance(arg_type, item_binding_point, section, options)
            if item_binding_point.has_new_value:
                collection.append(item_binding_point.value)
        except Exception as ex:
            logging.error(msg="", exc_info=ex)


def bind_set(collection: set, config: IConfiguration, options: BinderOptions, klass_type: type):
    arg_type = get_collection_arg_type(klass_type)
    for section in config.get_children():
        try:
            item_binding_point = BindingPoint()
            bind_instance(arg_type, item_binding_point, section, options)
            if item_binding_point.has_new_value:
                collection.add(item_binding_point.value)
        except Exception as ex:
            logging.error(msg="", exc_info=ex)


def bind_dictionary(dict_instance: dict, config: IConfiguration, options: BinderOptions, klass_type: type):
    assert isinstance(klass_type, GenericAlias)
    key_type = klass_type.__args__[0]
    value_type = klass_type.__args__[1]
    if not _is_valid_type_for_key_dict(key_type):
        raise ValueError("dictionary key type should be int or string")
    for child in config.get_children():
        try:
            key = get_key_value(child.key, key_type)

            def initial_value_provider_func():
                if key in dict_instance:
                    return dict_instance[key]
                return None

            value_binding_point = BindingPoint(initial_value_provider_func, False)
            bind_instance(value_type, value_binding_point, child, options)
            if value_binding_point.has_new_value:
                dict_instance[key] = value_binding_point.value
        except Exception as ex:
            logging.error("", ex)


def get_key_value(key: str, key_type: type):
    if issubclass(key_type, str):
        return key
    if issubclass(key_type, enum.Enum):
        for mem in key_type._member_map_:
            if mem.lower() == key.lower():
                return key_type[mem]
        raise ValueError(f"invalid enum value  key  {key} for type {key_type}")
    return key_type(key)


def _is_valid_type_for_key_dict(klass_type: type):
    return klass_type == int or \
        klass_type == str or \
        issubclass(klass_type, enum.Enum)



from multipledispatch import dispatch

from pyconfig_extension import guard
from pyconfig_extension.abstracts.iconfiguration import IConfigurationSection
from pyconfig_extension.abstracts.iconfiguration_root import IConfigurationRoot
import pyconfig_extension.abstracts.configuration_path as path_conf
import pyconfig_extension.core._internal_configuration_root_extensions as root_ext


class ConfigurationSection(IConfigurationSection):

    def __init__(self, root: IConfigurationRoot, path: str):
        guard.check_none(root)
        guard.check_none(path)

        self.__root = root
        self.__path = path
        self.__key = None

    @property
    def key(self) -> str:
        if self.__key is None:
            self.__key = path_conf.get_section_key(self.__path)
        return self.__key

    @property
    def path(self) -> str:
        return self.__path

    @property
    def value(self) -> str | None:
        return self.__root[self.path]

    @value.setter
    def value(self, val: str):
        self.__root[self.path] = val

    @dispatch(str)
    def __getitem__(self, key: str) -> str | None:
        root_key = path_conf.combine(self.path, key)
        return self.__root[root_key]

    def __setitem__(self, key: str, value: str | None):
        root_key = path_conf.combine(self.path, key)
        self.__root[root_key] = value

    def get_section(self, key: str) -> IConfigurationSection:
        root_key = path_conf.combine(self.path, key)
        return self.__root.get_section(root_key)

    def get_children(self) -> list[IConfigurationSection]:
        return root_ext.get_children_implementation(self.__root, self.path)

    def __str__(self):
        return f"path ={self.path} ,key ={self.key} , value= {self.value}"





from multipledispatch import dispatch

from pyconfig_extension import guard
from pyconfig_extension.abstracts.iconfiguration import IConfiguration
from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.core.chained_configuration_source import ChainedConfigurationSource
from pyconfig_extension.core.memory_configuration_source import MemoryConfigurationSource


def add_configuration(configuration_builder: IConfigurationBuilder, config: IConfiguration,
                      should_dispose_configuration: bool = False):
    guard.check_none(configuration_builder)
    guard.check_none(config)

    source = ChainedConfigurationSource(config, should_dispose_configuration)
    configuration_builder.add(source)
    return configuration_builder


@dispatch(IConfigurationBuilder)
def add_in_memory_collection(builder: IConfigurationBuilder) -> IConfigurationBuilder:
    builder.add(MemoryConfigurationSource())
    return builder


@dispatch(IConfigurationBuilder, dict)
def add_in_memory_collection(builder: IConfigurationBuilder, internal_data: dict[str, str]) -> IConfigurationBuilder:
    builder.add(MemoryConfigurationSource(internal_data))
    return builder



from abc import abstractmethod, ABC

from multipledispatch import dispatch

from pyconfig_extension.abstracts.stream_configuration_source import StreamConfigurationSource
from pyconfig_extension.core.configuration_provider import ConfigurationProvider


class StreamConfigurationProvider(ConfigurationProvider, ABC):

    def __init__(self, source: StreamConfigurationSource):
        super().__init__()
        self._source = source
        self._loaded = False

    def load(self):
        if self._loaded:
            raise Exception()
        self.load_stream(self._source.stream)
        self._loaded = True

    @abstractmethod
    def load_stream(self, stream):
        raise NotImplementedError



from abc import ABC, abstractmethod


class IConfigurationProvider(ABC):
    """
    Provides configuration key/values for an application.
    """

    @abstractmethod
    def try_get(self, key: str) -> tuple[bool, str | None]:
        raise NotImplementedError

    @abstractmethod
    def set(self, key: str, value: str):
        """
        Sets a configuration value for the specified key.
        :param key: The key.
        :param value: The value.
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def load(self):
        """
        Loads configuration values from the source represented by this 'IConfigurationProvider'
        """
        raise NotImplementedError

    @abstractmethod
    def get_child_keys(self, earlier_keys: list[str], parent_path: str | None) -> list[str]:
        """
        Returns the immediate descendant configuration keys for a given parent path based on this
        'IConfigurationProvider's data and the set of keys returned by all the preceding
        :param earlier_keys: The child keys returned by the preceding providers for the same parent path.
        :param parent_path: The parent path.
        :return: The child keys.
        """
        raise NotImplementedError

    def get(self, key: str):
        checked, value = self.try_get(key)
        if checked:
            return value
        raise Exception(f"key not found key={key}")



from multipledispatch import dispatch

from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.command_line.source import CommandLineConfigurationSource


@dispatch(IConfigurationBuilder, list)
def add_command_line(builder: IConfigurationBuilder, args: list[str]) -> IConfigurationBuilder:
    return add_command_line(builder, args, None)


@dispatch(IConfigurationBuilder, list, dict)
def add_command_line(builder: IConfigurationBuilder, args: list[str],
                     switch_mappings: dict[str, str] | None) -> IConfigurationBuilder:
    source = CommandLineConfigurationSource(args, switch_mappings)
    return builder.add(source)



from pyconfig_extension.file.file_configuration_provider import FileConfigurationProvider
from pyconfig_extension.ini.ini_configuration_source import IniConfigurationSource
from pyconfig_extension.ini.ini_stream_configuration_provider import IniStreamConfigurationProvider


class IniConfigurationProvider(FileConfigurationProvider):
    def __init__(self, source: IniConfigurationSource):
        super().__init__(source)

    def load_stream(self, stream):
        self._data = IniStreamConfigurationProvider.read(stream)



from pyconfig_extension.abstracts.configuration_path import KEY_DELIMITER
from pyconfig_extension.core.ordinal_ignore_case_dict import OrdinalIgnoreCaseDict
import json5


class JsonConfigurationFileParser:

    def __init__(self):
        self._path: list[str] = []
        self._data = OrdinalIgnoreCaseDict()

    @staticmethod
    def parse(stream) -> OrdinalIgnoreCaseDict:
        parser = JsonConfigurationFileParser()
        return parser.parse_stream(stream)

    def parse_stream(self, stream) -> OrdinalIgnoreCaseDict:
        root = json5.loads(stream)
        self.__visit_object_element(root)
        return self._data

    def __visit_object_element(self, element):
        is_empty = True
        for key in element:
            is_empty = False
            self.__enter_context(key)
            self.__visit_value(element[key])
            self.__exit_context()
        self.__set_null_if_element_is_empty(is_empty)

    def __set_null_if_element_is_empty(self, is_empty: bool):
        if is_empty and len(self._path) > 0:
            key = self._path[-1]
            self._data[key] = None

    def __visit_value(self, value):
        if isinstance(value, str) or isinstance(value, bool) or isinstance(value, int):
            key = self._path[-1]
            if key in self._data:
                raise ValueError(f"A duplicate key '{key}' was found.")
            self._data[key] = str(value)
        elif value is None:
            key = self._path[-1]
            self._data[key] = ""
        elif isinstance(value, dict):
            self.__visit_object_element(value)
        elif isinstance(value, list):
            self.__visit_array_elements(value)
        else:
            raise ValueError(f"Unsupported JSON token '{type(value)}' was found.")

    def __visit_array_elements(self, value: list):
        index = 0
        for element in value:
            self.__enter_context(str(index))
            self.__visit_value(element)
            self.__exit_context()
            index += 1

        self.__set_null_if_element_is_empty(is_empty=index == 0)

    def __enter_context(self, context: str):
        path = self._path[-1] + KEY_DELIMITER + context if len(self._path) > 0 else context
        self._path.append(path)

    def __exit_context(self):
        self._path.pop()



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.stream_configuration_source import StreamConfigurationSource


class IniStreamConfigurationSource(StreamConfigurationSource):
    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.ini.ini_stream_configuration_provider import IniStreamConfigurationProvider
        return IniStreamConfigurationProvider(self)



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.file.file_configuration_source import FileConfigurationSource


class JsonConfigurationSource(FileConfigurationSource):
    def __init__(self, stream=None, path: str = None, optional: bool = False):
        super().__init__(stream, path, optional)

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        self.ensure_defaults(builder)
        from pyconfig_extension.json.json_configuration_provider import JsonConfigurationProvider
        return JsonConfigurationProvider(self)



from typing import Generator, Callable

from pyconfig_extension.abstracts.iconfiguration import IConfiguration ,IConfigurationSection
from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder


def as_generator(self: IConfiguration, make_paths_relative: bool = False) -> Generator[tuple[str, str], None, None]:
    stack: list[IConfiguration] = [self]

    prefix_length = 0
    if make_paths_relative and isinstance(self, IConfigurationSection):
        prefix_length = len(self.path) + 1

    while len(stack) > 0:
        config: IConfiguration = stack.pop()

        if isinstance(config, IConfigurationSection) and (not make_paths_relative or config != self):
            yield config.path[prefix_length:], config.value

        for child in config.get_children():
            stack.append(child)


def add(builder: IConfigurationBuilder, config_type: type,
        configure_source: Callable = None) -> IConfigurationBuilder:
    source = config_type()
    if configure_source is not None:
        configure_source(source)
    return builder.add(source)



import math

KEY_DELIMITER = ":"


def __skip_ahead_on_delimiter(inp: str):
    for i in range(len(inp)):
        if inp[i] != KEY_DELIMITER:
            return inp[i:]
    return ""


def is_integer(s: str) -> bool:
    """
    Checks if a string represents an integer value.

    Returns True if the string is an integer, False otherwise.
    """
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()


def __just_compare(a: str | None, b: str | None) -> int:
    a_is_int = is_integer(a)
    b_is_int = is_integer(b)

    if not a_is_int and not b_is_int:
        if a.lower() < b.lower():
            return -1
        elif a.lower() > b.lower():
            return 1
        return 0
    elif a_is_int and b_is_int:
        a = int(a)
        b = int(b)
        if a < b:
            return -1
        elif a > b:
            return 1
        else:
            return 0
    else:

        if a_is_int:
            return -1
        return 1


def __delimiter_index(val: str):
    try:
        return val.index(KEY_DELIMITER)
    except:
        return -1


def compare(x: str | None, y: str | None) -> int:
    x = "" if x is None else x
    y = "" if y is None else y
    x = __skip_ahead_on_delimiter(x)
    y = __skip_ahead_on_delimiter(y)

    while len(x) != 0 and len(y) != 0:
        x_delimiter_index = __delimiter_index(x)
        y_delimiter_index = __delimiter_index(y)

        a = x if x_delimiter_index == -1 else x[0:x_delimiter_index]
        b = y if y_delimiter_index == -1 else y[0:y_delimiter_index]
        compare_result = __just_compare(a, b)

        if compare_result != 0:
            return compare_result

        x = "" if x_delimiter_index == -1 else __skip_ahead_on_delimiter(x[x_delimiter_index + 1:])
        y = "" if y_delimiter_index == -1 else __skip_ahead_on_delimiter(y[y_delimiter_index + 1:])

    if len(x) == 0:
        if len(y) == 0:
            return 0
        return -1
    return 1



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.file.file_configuration_source import FileConfigurationSource


class IniConfigurationSource(FileConfigurationSource):

    def build(self, builder: IConfigurationBuilder) -> IConfigurationProvider:
        from pyconfig_extension.ini.ini_configuration_provider import IniConfigurationProvider
        return IniConfigurationProvider(self)




from pyconfig_extension.file.file_configuration_provider import FileConfigurationProvider
from pyconfig_extension.file.file_configuration_source import FileConfigurationSource
from pyconfig_extension.json.json_configuration_file_parser import JsonConfigurationFileParser


class JsonConfigurationProvider(FileConfigurationProvider):

    def __init__(self, source: FileConfigurationSource):
        super().__init__(source)

    def load_stream(self, stream):
        self._data = JsonConfigurationFileParser.parse(stream)



from pyconfig_extension.abstracts.configuration_path import KEY_DELIMITER
from pyconfig_extension.core.configuration_provider import ConfigurationProvider
from pyconfig_extension.core.ordinal_ignore_case_dict import OrdinalIgnoreCaseDict

MY_SQL_SERVER_PREFIX = "MYSQLCONNSTR_"
SQL_AZURE_SERVER_PREFIX = "SQLAZURECONNSTR_"
SQL_SERVER_PREFIX = "SQLCONNSTR_"
CUSTOM_CONNECTION_STRING_PREFIX = "CUSTOMCONNSTR_"


class EnvironmentVariablesConfigurationProvider(ConfigurationProvider):
    def __init__(self, prefix: str | None = None):
        super().__init__()
        self.__prefix = "" if prefix is None else prefix
        self.__normal_prefix = self.__normalize(self.__prefix)

    def load_data(self, data: dict[str, str]):
        res = OrdinalIgnoreCaseDict()
        for key in data:
            value = data[key]
            self.__add_if_normalized_key_matches_prefix(res, self.__normalize(key), value)
        self._data = res

    def __add_if_normalized_key_matches_prefix(self, data: OrdinalIgnoreCaseDict, normalized_key: str, value: str):
        if normalized_key.lower().startswith(self.__normal_prefix.lower()):
            data[normalized_key[len(self.__normal_prefix):]] = value

    @staticmethod
    def __normalize(key: str):
        return key.replace("__", KEY_DELIMITER)

    def __str__(self):
        return f"{type(self).__name__} Prefix: '{self.__prefix}'"



from io import StringIO

from pyconfig_extension.abstracts.configuration_path import KEY_DELIMITER
from pyconfig_extension.abstracts.stream_configuration_provider import StreamConfigurationProvider
from pyconfig_extension.core.ordinal_ignore_case_dict import OrdinalIgnoreCaseDict
from pyconfig_extension.ini.ini_stream_configuration_source import IniStreamConfigurationSource
import os


def get_index(inp: str, delim: str) -> int:
    try:
        return inp.index(delim)
    except:
        return -1


class IniStreamConfigurationProvider(StreamConfigurationProvider):

    def __init__(self, source: IniStreamConfigurationSource):
        super().__init__(source)

    def load_stream(self, stream):
        self._data = self.read(stream)

    @staticmethod
    def read(stream: str) -> OrdinalIgnoreCaseDict:
        result = OrdinalIgnoreCaseDict()
        section_prefix = ""
        lines = stream.splitlines()
        for line in lines:
            raw_line = line.strip()
            # ignore blank lines
            if len(raw_line) == 0:
                continue
            # ignore comments
            if raw_line[0] in ';#/':
                continue
            # [Section:header]
            if raw_line[0] == '[' and raw_line[-1] == ']':
                # remove the brackets
                section_prefix = raw_line[1:-1].strip() + KEY_DELIMITER
                continue

            # key = value or "value"
            seperator = get_index(raw_line, "=")
            if seperator < 0:
                raise Exception(f"Unrecognized line format: '{line}'.")
            key = section_prefix + raw_line[: seperator].strip()
            value = raw_line[seperator + 1:].strip()
            if len(value) > 0 and value[0] == '"' and value[-1] == '"':
                value = value[1:-1]

            if key in result:
                raise Exception(f"A duplicate key '{key}' was found.")
            result[key] = value
        return result



from abc import ABC, abstractmethod

from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.iconfiguration import IConfiguration


class IConfigurationRoot(IConfiguration, ABC):

    @abstractmethod
    def reload(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def providers(self) -> list[IConfigurationProvider]:
        raise NotImplementedError



from pyconfig_extension.abstracts.iconfiguration_builder import IConfigurationBuilder, IConfigurationSource
from pyconfig_extension.abstracts.iconfiguration_root import IConfigurationRoot
import pyconfig_extension.guard as guard


class ConfigurationBuilder(IConfigurationBuilder):
    def __init__(self):
        self.__properties: dict[str, object] = {}
        self.__sources: list[IConfigurationSource] = []

    @property
    def properties(self) -> dict[str, object]:
        return self.__properties

    @property
    def sources(self) -> list[IConfigurationSource]:
        return self.__sources

    def add(self, source: IConfigurationSource) -> IConfigurationBuilder:
        guard.check_none(source)

        self.sources.append(source)
        return self

    def build(self) -> IConfigurationRoot:
        providers = []
        for source in self.sources:
            provider = source.build(self)
            providers.append(provider)
        from pyconfig_extension.core.configuration_root import ConfigurationRoot
        return ConfigurationRoot(providers)



from multipledispatch import dispatch

from pyconfig_extension import guard
from pyconfig_extension.abstracts.iconfiguration import IConfiguration
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.core.chained_configuration_source import ChainedConfigurationSource


class ChainedConfigurationProvider(IConfigurationProvider):

    def __init__(self, source: ChainedConfigurationSource):
        guard.check_none(source.configuration)
        self._config = source.configuration
        self._should_exit_config = source.should_exit_configuration

    @property
    def configuration(self) -> IConfiguration:
        """
        Gets the chained configuration.
        """
        return self._config

    def try_get(self, key: str) -> tuple[bool, str | None]:
        value = self._config[key]
        return value is not None, value

    def set(self, key: str, value: str):
        self._config[key] = value

    def load(self):
        pass

    def get_child_keys(self, earlier_keys: list[str], parent_path: str | None) -> list[str]:
        section: IConfiguration = self._config if parent_path is None else self._config.get_section(parent_path)

        keys: list[str] = []
        for child in section.get_children():
            keys.append(child.key)

        keys.extend(earlier_keys)
        keys.sort(key=lambda x: x.lower())
        return keys

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._should_exit_config:
            if hasattr(self._config, "__exit__"):
                self._config.__exit__(exc_type, exc_val, exc_tb)



from pyconfig_extension.abstracts.iconfiguration import IConfigurationSection
from pyconfig_extension.abstracts.iconfiguration_provider import IConfigurationProvider
from pyconfig_extension.abstracts.iconfiguration_root import IConfigurationRoot
import pyconfig_extension.core._internal_configuration_root_extensions as ext
from pyconfig_extension.core.configuration_section import ConfigurationSection


class ConfigurationRoot(IConfigurationRoot):
    def __init__(self, providers: list[IConfigurationProvider]):
        self.__providers = providers
        for p in providers:
            p.load()

    def reload(self):
        for provider in self.__providers:
            provider.load()

    @property
    def providers(self) -> list[IConfigurationProvider]:
        return self.__providers

    def __getitem__(self, item: str) -> str:
        return self.get_configuration(self.providers, item)

    def __setitem__(self, key: str, value: str):
        self.set_configuration(self.providers, key, value)

    def get_section(self, key) -> IConfigurationSection:
        return ConfigurationSection(self, key)

    def get_children(self) -> list[IConfigurationSection]:
        return ext.get_children_implementation(self, None)

    @staticmethod
    def get_configuration(providers: list[IConfigurationProvider], key: str) -> str | None:
        for provider in providers[::-1]:
            checked, value = provider.try_get(key)
            if checked:
                return value
        return None

    @staticmethod
    def set_configuration(providers: list[IConfigurationProvider], key: str, value: str | None):
        if len(providers) == 0:
            raise Exception("There is no source for set configuration")

        for provider in providers:
            provider.set(key, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for provider in self.__providers:
            if hasattr(provider, "__exit__"):
                provider.__exit__(exc_type, exc_val, exc_tb)



class OrdinalIgnoreCaseDict:
    def __init__(self):
        self.__data = {}
        self.__low_key_data: dict[str, str] = {}

    def __setitem__(self, key: str, value):
        if not isinstance(key, str):
            ValueError("key should be string")
        dict_key = self.__dict_key(key)
        if dict_key not in self.__data:
            self.__data[dict_key] = value
            self.__low_key_data[key.lower()] = key
        else:
            del self.__data[dict_key]
            self.__data[key] = value
            self.__low_key_data[key.lower()] = key

    def __getitem__(self, key: str):
        if not isinstance(key, str):
            ValueError("key should be string")
        dict_key = self.__dict_key(key)
        return self.__data[dict_key]

    def __delitem__(self, key: str):
        if not isinstance(key, str):
            ValueError("key should be string")
        dict_key = self.__dict_key(key)
        self.__data.__delitem__(dict_key)
        self.__low_key_data.__delitem__(key.lower())

    def __len__(self):
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key: str):
        if not isinstance(key, str):
            ValueError("key should be string")
        dict_key = self.__dict_key(key)
        return self.__data.__contains__(dict_key)

    def try_get(self, key: str) -> tuple[bool, object | None]:
        dict_key = self.__dict_key(key)
        if dict_key in self.__data:
            return True, self.__data[dict_key]
        return False, None

    def __dict_key(self, key: str):
        dict_key = key
        if key.lower() in self.__low_key_data:
            dict_key = self.__low_key_data[key.lower()]
        return dict_key
